from typing import cast
from unittest.mock import Mock, patch

import boto3
from google.cloud import batch_v1, compute_v1
from google.protobuf.json_format import MessageToDict  # type: ignore

from redun import File, task
from redun.config import Config
from redun.executors.gcp_batch import GCPBatchExecutor
from redun.expression import TaskExpression
from redun.scheduler import Job, Traceback
from redun.scripting import script, script_task
from redun.tests.utils import (
    get_filesystem_class_mock,
    mock_s3,
    mock_scheduler,
    use_tempdir,
    wait_until,
)
from redun.utils import pickle_dumps

PROJECT = "project"
IMAGE = "gcr.io/gcp-project/redun"
REGION = "region"
GCS_SCRATCH_PREFIX = "gs://example-bucket/redun"


def mock_executor(scheduler, debug=False, code_package=True) -> GCPBatchExecutor:
    """
    Returns an GCPBatchExecutor with GCP API mocks.
    """
    # We use a mocked s3 backend to simulate gcs.
    client = boto3.client("s3", region_name="us-east-1")
    client.create_bucket(Bucket="example-bucket")

    # Setup executor.
    config = Config(
        {
            "gcp_batch": {
                "gcs_scratch": GCS_SCRATCH_PREFIX,
                "project": PROJECT,
                "region": PROJECT,
                "image": IMAGE,
                "job_monitor_interval": 0.05,
                "job_stale_time": 0.01,
                "code_package": code_package,
                "debug": debug,
                "min_array_size": 2,
            }
        }
    )

    executor = GCPBatchExecutor("batch", scheduler, config["gcp_batch"])
    executor.start()

    def executor_start():
        executor.is_running = True
        # Prevent monitor thread from running.
        executor._thread = Mock()

    executor._start = executor_start  # type: ignore

    return executor


@task
def task1(x: int) -> int:
    return x


@use_tempdir
@mock_s3
@patch("redun.file.get_filesystem_class", get_filesystem_class_mock)
@patch("redun.executors.gcp_utils.get_gcp_compute_client")
@patch("redun.executors.gcp_utils.get_compute_machine_type")
@patch("redun.executors.gcp_utils.get_gcp_batch_client")
@patch("redun.executors.gcp_utils.list_jobs")
@patch("redun.executors.gcp_utils.get_task")
def test_executor(
    get_task_mock: Mock,
    list_jobs_mock: Mock,
    get_gcp_batch_client_mock: Mock,
    get_compute_machine_type_mock: Mock,
    get_gcp_compute_client_mock: Mock,
) -> None:
    """
    GCPBatchExecutor should run jobs.
    """
    scheduler = mock_scheduler()
    executor = mock_executor(scheduler)
    client = get_gcp_batch_client_mock()

    # Prepare API mocks for job submission.
    batch_job_id = "123"
    batch_job = batch_v1.Job(task_groups=[batch_v1.TaskGroup(name=batch_job_id)])
    client.create_job.return_value = batch_job
    list_jobs_mock.return_value = []
    get_task_mock.return_value = batch_v1.Task(
        name=f"{batch_job_id}/tasks/0",
        status=batch_v1.TaskStatus(state=batch_v1.TaskStatus.State.SUCCEEDED),
    )
    get_compute_machine_type_mock.return_value = compute_v1.types.MachineType(
        memory_mb=16384,
        guest_cpus=2,
    )

    # Create and submit a job.
    expr = cast(TaskExpression[int], task1(10))
    job = Job(task1, expr)
    job.eval_hash = "eval_hash"
    job.args = ((10,), {})
    executor.submit(job)

    # Let job arrayer submit job.
    wait_until(lambda: executor.arrayer.num_pending == 0)

    # Ensure job was submitted correctly to Google Batch.
    scratch_dir = executor.gcs_scratch_prefix
    assert executor.code_file
    assert MessageToDict(client.create_job.call_args[0][0]._pb) == {
        "job": {
            "allocationPolicy": {"instances": [{"policy": {"machineType": "e2-standard-4"}}]},
            "labels": {"redun_hash": "eval_hash", "redun_job_type": "container"},
            "logsPolicy": {"destination": "CLOUD_LOGGING"},
            "priority": "30",
            "taskGroups": [
                {
                    "taskCount": "1",
                    "taskSpec": {
                        "computeResource": {"cpuMilli": "2000", "memoryMib": "16384"},
                        "maxRetryCount": 2,
                        "runnables": [
                            {
                                "container": {
                                    "commands": [
                                        "redun",
                                        "--check-version",
                                        ">=0.4.1",
                                        "oneshot",
                                        "redun.tests.test_gcp_batch",
                                        "--code",
                                        executor.code_file.path,
                                        "--input",
                                        "gs://example-bucket/redun/jobs/eval_hash/input",
                                        "--output",
                                        "gs://example-bucket/redun/jobs/eval_hash/output",
                                        "--error",
                                        "gs://example-bucket/redun/jobs/eval_hash/error",
                                        "task1",
                                    ],
                                    "imageUri": "gcr.io/gcp-project/redun",
                                }
                            }
                        ],
                    },
                }
            ],
        },
        "jobId": f"redun-{job.id}",
        "parent": "projects/project/locations/project",
    }

    # Simulate output file created by job.
    output_file = File(f"{scratch_dir}/jobs/eval_hash/output")
    output_file.write(pickle_dumps(task1.func(10)), mode="wb")

    # Manually run monitor logic.
    executor._monitor()

    # Ensure job returns result to scheduler.
    assert scheduler.job_results[job.id] == 10

    # Prepare API mocks for job submission of a failed job.
    batch_job_id = "456"
    batch_job = batch_v1.Job(task_groups=[batch_v1.TaskGroup(name=batch_job_id)])
    client.create_job.return_value = batch_job
    list_jobs_mock.return_value = []
    get_task_mock.return_value = batch_v1.Task(
        name=f"{batch_job_id}/tasks/0",
        status=batch_v1.TaskStatus(state=batch_v1.TaskStatus.State.FAILED),
    )

    # Create and submit a job.
    expr = cast(TaskExpression[int], task1(11))
    job = Job(task1, expr)
    job.eval_hash = "eval_hash2"
    job.args = ((11,), {})
    executor.submit(job)

    # Let job arrayer submit job.
    wait_until(lambda: executor.arrayer.num_pending == 0)

    # Simulate the job failing.
    error = ValueError("Boom")
    error_traceback = Traceback.from_error(error)
    error_file = File(f"{scratch_dir}/jobs/eval_hash2/error")
    error_file.write(pickle_dumps((error, error_traceback)), mode="wb")

    # Manually run monitor thread.
    executor._monitor()

    # Ensure job rejected to scheduler.
    assert repr(scheduler.job_errors[job.id]) == "ValueError('Boom')"


@use_tempdir
@mock_s3
@patch("redun.file.get_filesystem_class", get_filesystem_class_mock)
@patch("redun.executors.gcp_utils.get_gcp_compute_client")
@patch("redun.executors.gcp_utils.get_compute_machine_type")
@patch("redun.executors.gcp_utils.get_gcp_batch_client")
@patch("redun.executors.gcp_utils.list_jobs")
@patch("redun.executors.gcp_utils.get_task")
def test_executor_array(
    get_task_mock: Mock,
    list_jobs_mock: Mock,
    get_gcp_batch_client_mock: Mock,
    get_compute_machine_type_mock: Mock,
    get_gcp_compute_client_mock: Mock,
) -> None:
    """
    GCPBatchExecutor should be able to submit array jobs.
    """
    scheduler = mock_scheduler()
    executor = mock_executor(scheduler)
    client = get_gcp_batch_client_mock()

    # Suppress inflight jobs check.
    list_jobs_mock.return_value = []

    # Prepare API mocks for job submission.
    client.create_job.return_value = batch_v1.Job(
        task_groups=[
            batch_v1.TaskGroup(
                name="123",
                task_count=2,
            )
        ]
    )

    def get_task(client, task_name):
        if task_name.endswith("0"):
            state = batch_v1.TaskStatus.State.SUCCEEDED
        else:
            state = batch_v1.TaskStatus.State.FAILED

        return batch_v1.Task(
            name=task_name,
            status=batch_v1.TaskStatus(state=state),
        )

    get_task_mock.side_effect = get_task
    get_compute_machine_type_mock.return_value = compute_v1.types.MachineType(
        memory_mb=16384,
        guest_cpus=2,
    )

    # Submit two jobs in order to trigger array submission.
    # Create and submit a job.
    expr = cast(TaskExpression[int], task1(10))
    job1 = Job(task1, expr)
    job1.eval_hash = "eval_hash"
    job1.args = ((10,), {})
    executor.submit(job1)

    # Create and submit a job.
    expr = cast(TaskExpression[int], task1(11))
    job2 = Job(task1, expr)
    job2.eval_hash = "eval_hash2"
    job2.args = ((11,), {})
    executor.submit(job2)

    # Let job arrayer submit job.
    wait_until(lambda: executor.arrayer.num_pending == 0)

    # Ensure job was submitted to correctly to Google Batch.
    scratch_dir = executor.gcs_scratch_prefix
    assert executor.code_file

    response = MessageToDict(client.create_job.call_args[0][0]._pb)
    array_uuid = response["jobId"].replace("redun-array-", "")
    assert response == {
        "job": {
            "allocationPolicy": {"instances": [{"policy": {"machineType": "e2-standard-4"}}]},
            "labels": {"redun_hash": array_uuid, "redun_job_type": "container"},
            "logsPolicy": {"destination": "CLOUD_LOGGING"},
            "priority": "30",
            "taskGroups": [
                {
                    "taskCount": "2",
                    "taskSpec": {
                        "computeResource": {"cpuMilli": "2000", "memoryMib": "16384"},
                        "maxRetryCount": 2,
                        "runnables": [
                            {
                                "container": {
                                    "commands": [
                                        "redun",
                                        "--check-version",
                                        ">=0.4.1",
                                        "oneshot",
                                        "redun.tests.test_gcp_batch",
                                        "--code",
                                        executor.code_file.path,
                                        "--array-job",
                                        "--input",
                                        f"gs://example-bucket/redun/array_jobs/{array_uuid}/input",  # noqa: E501
                                        "--output",
                                        f"gs://example-bucket/redun/array_jobs/{array_uuid}/output",  # noqa: E501
                                        "--error",
                                        f"gs://example-bucket/redun/array_jobs/{array_uuid}/error",  # noqa: E501
                                        "task1",
                                    ],
                                    "imageUri": "gcr.io/gcp-project/redun",
                                }
                            }
                        ],
                    },
                }
            ],
        },
        "jobId": f"redun-array-{array_uuid}",
        "parent": "projects/project/locations/project",
    }

    # Simulate output file created by job1.
    output_file = File(f"{scratch_dir}/jobs/eval_hash/output")
    output_file.write(pickle_dumps(task1.func(10)), mode="wb")

    # Simulate job2 failing.
    error = ValueError("Boom")
    error_traceback = Traceback.from_error(error)
    error_file = File(f"{scratch_dir}/jobs/eval_hash2/error")
    error_file.write(pickle_dumps((error, error_traceback)), mode="wb")

    # Manually run monitor logic.
    executor._monitor()

    # Ensure job returns results and errors to scheduler.
    assert scheduler.job_results[job1.id] == 10
    assert repr(scheduler.job_errors[job2.id]) == "ValueError('Boom')"


@use_tempdir
@mock_s3
@patch("redun.file.get_filesystem_class", get_filesystem_class_mock)
@patch("redun.executors.gcp_utils.get_gcp_compute_client")
@patch("redun.executors.gcp_utils.get_compute_machine_type")
@patch("redun.executors.gcp_utils.get_gcp_batch_client")
@patch("redun.executors.gcp_utils.list_jobs")
@patch("redun.executors.gcp_utils.get_task")
def test_executor_script(
    get_task_mock: Mock,
    list_jobs_mock: Mock,
    get_gcp_batch_client_mock: Mock,
    get_compute_machine_type_mock: Mock,
    get_gcp_compute_client_mock: Mock,
) -> None:
    """
    GCPBatchExecutor should run script jobs.
    """
    scheduler = mock_scheduler()
    executor = mock_executor(scheduler)
    client = get_gcp_batch_client_mock()

    # Prepare API mocks for job submission.
    batch_job_id = "123"
    batch_job = batch_v1.Job(task_groups=[batch_v1.TaskGroup(name=batch_job_id)])
    client.create_job.return_value = batch_job
    list_jobs_mock.return_value = []
    get_task_mock.return_value = batch_v1.Task(
        name=f"{batch_job_id}/tasks/0",
        status=batch_v1.TaskStatus(state=batch_v1.TaskStatus.State.SUCCEEDED),
    )
    get_compute_machine_type_mock.return_value = compute_v1.types.MachineType(
        memory_mb=16384,
        guest_cpus=2,
    )
    # Create and submit a script job.
    # Simulate the call to script_task().
    expr = script(
        "cat input > output",
        inputs=[File("gs://example-bucket/input").stage("input")],
        output=File("gs://example-bucket/output").stage("output"),
    )
    command = expr.args[0]
    expr = script_task(command)
    job = Job(script_task, expr)
    job.eval_hash = "eval_hash"
    job.args = (expr.args, expr.kwargs)
    executor.submit_script(job)

    # Let job arrayer submit job.
    wait_until(lambda: executor.arrayer.num_pending == 0)

    # Ensure job was submitted correctly to Google Batch.
    assert MessageToDict(client.create_job.call_args[0][0]._pb) == {
        "job": {
            "allocationPolicy": {"instances": [{"policy": {"machineType": "e2-standard-4"}}]},
            "labels": {"redun_hash": "eval_hash", "redun_job_type": "script"},
            "logsPolicy": {"destination": "CLOUD_LOGGING"},
            "priority": "30",
            "taskGroups": [
                {
                    "taskCount": "1",
                    "taskSpec": {
                        "computeResource": {"cpuMilli": "2000", "memoryMib": "16384"},
                        "maxRetryCount": 2,
                        "runnables": [
                            {
                                "container": {
                                    "commands": [
                                        "bash",
                                        "/mnt/disks/example-bucket/redun/jobs/eval_hash/.redun_job.sh",  # noqa: E501
                                    ],
                                    "imageUri": "gcr.io/gcp-project/redun",
                                    "volumes": [
                                        "/mnt/disks/example-bucket:/mnt/disks/example-bucket"
                                    ],
                                }
                            }
                        ],
                        "volumes": [
                            {
                                "gcs": {"remotePath": "example-bucket"},
                                "mountPath": "/mnt/disks/example-bucket",
                            }
                        ],
                    },
                }
            ],
        },
        "jobId": f"redun-{job.id}",
        "parent": "projects/project/locations/project",
    }

    # Simulate output file created by job.
    scratch_dir = executor.gcs_scratch_prefix
    output_file = File(f"{scratch_dir}/jobs/eval_hash/output")
    output_file.write("done")

    # Manually run monitor logic.
    executor._monitor()

    # Ensure job returns result to scheduler.
    assert scheduler.job_results[job.id] == [0, b"done"]

    # Prepare API mocks for job submission.
    batch_job_id = "456"
    batch_job = batch_v1.Job(task_groups=[batch_v1.TaskGroup(name=batch_job_id)])
    client.create_job.return_value = batch_job
    list_jobs_mock.return_value = []
    get_task_mock.return_value = batch_v1.Task(
        name=f"{batch_job_id}/tasks/0",
        status=batch_v1.TaskStatus(state=batch_v1.TaskStatus.State.FAILED),
    )

    # Submit a failing script task.
    job2 = Job(script_task, expr)
    job2.eval_hash = "eval_hash2"
    job2.args = (expr.args, expr.kwargs)
    executor.submit_script(job2)

    # Let job arrayer submit job.
    wait_until(lambda: executor.arrayer.num_pending == 0)

    # Simulate output file created by job.
    error_file = File(f"{scratch_dir}/jobs/eval_hash2/error")
    error_file.write("boom")

    # Manually run monitor logic.
    executor._monitor()

    # Ensure job returns result to scheduler.
    assert repr(scheduler.job_errors[job2.id]) == "ScriptError('Last line: boom')"



import re
from typing import Any, Dict, List, Tuple

import pygraphviz as pgv
from pygraphviz import AGraph

from redun.backends.db import CallNode, Execution, Job, RedunBackendDb, Value
from redun.backends.db.dataflow import ArgumentValue, CallNodeValue, walk_dataflow
from redun.backends.db.query import CallGraphQuery
from redun.scheduler import ErrorValue, Scheduler

JOB_COLOR = "#8FE0AC"
VALUE_COLOR = "#FF8484"
CALL_NODE_COLOR = "#B5D3E7"
SUBVALUE_EDGE_COLOR = "#9E9E9E"
DEFAULT_COLOR = ""
MAX_LABEL_LEN = 100
HALF_LABEL_LEN = MAX_LABEL_LEN // 2
SEPARATOR = "\n...\n"


def truncate(string: str) -> str:
    """
    Truncates a node label.
    """
    return string[:HALF_LABEL_LEN] + SEPARATOR + string[-HALF_LABEL_LEN:]


def clean_repr(value: Value) -> str:
    """
    Produces a representation of a value without the hash=<hash-id> section.
    """
    hash_regex = ", hash=.+?(?=\\))"
    return re.sub(hash_regex, "", repr(value.value_parsed))


def populate_valid_nodes(scheduler: Scheduler, exec_id: str, props: Dict[str, Any]) -> None:
    """
    Collects all the CallNodes associated with the provided execution id into a property.

    Used to validate CallNodes when accessed via parent-child relationships, which are not
    necessarily internal to an execution.
    """
    assert isinstance(scheduler.backend, RedunBackendDb)

    query = CallGraphQuery(scheduler.backend.session)
    call_nodes_query = query.filter_execution_ids([exec_id]).filter_types({"CallNode"})

    valid_call_nodes = [call_node.call_hash for call_node in call_nodes_query.all()]
    props["valid_call_nodes"] = valid_call_nodes


def init_graph(direction: str) -> AGraph:
    """
    Initializes a graph with the desired attributes.
    """
    graph = pgv.AGraph(strict=False, directed=True)
    graph.node_attr.update(
        {
            "shape": "box",
            "style": "filled",
            "fontname": "helvetica",
        }
    )
    graph.graph_attr.update(
        {
            "clusterrank": "local",
            "rankdir": direction,
            "splines": "polyline",
            "ranksep": 2,
            "fontname": "helvetica",
        }
    )
    return graph


def add_value(
    value: Value, upstream: Any, graph: AGraph, props: Dict[str, Any]
) -> List[Tuple[Value, Any]]:
    """
    Adds a Value and its subvalues to the provided graph.

    To give different values that may have the same parsed representation a unique identifier
    in the graph, we package the value with its upstream lineage. The upstream argument will either
    be a list of CallNodes or the arg hash if the value is top level.
    """
    color = VALUE_COLOR if isinstance(value.value_parsed, ErrorValue) else DEFAULT_COLOR
    if not props["dataflow"] or not value.children:
        hash_str = f" {value.value_hash[:8]}" if props["hash"] else ""
        value_str = f"Value{hash_str}\n{clean_repr(value)}"
        if len(value_str) > MAX_LABEL_LEN and not props.get("no_truncation", False):
            value_str = truncate(value_str)
        graph.add_node((value, upstream), fillcolor=color, label=value_str)
        return [(value, upstream)]
    else:
        values = []
        # This should go down to the leaf subvalues, as desired
        for arg_child in value.children:
            hash_str = f" {arg_child.value_hash[:8]}" if props["hash"] else ""
            value_str = f"Value{hash_str}\n{clean_repr(arg_child)}"
            if len(value_str) > MAX_LABEL_LEN and not props.get("no_truncation", False):
                value_str = truncate(value_str)
            graph.add_node((arg_child, upstream), label=value_str)
            values.append((arg_child, upstream))
        return values


def add_call_node(call_node: CallNode, graph: AGraph, props: Dict[str, Any]) -> None:
    """
    Adds a CallNode to the provided graph.
    """
    hash_str = f" {call_node.call_hash[:8]}" if props["hash"] else ""
    call_node_str = f"CallNode{hash_str}\n{call_node.task.fullname}"
    label = props.get("job_str", "") + call_node_str
    graph.add_node(call_node, fillcolor=CALL_NODE_COLOR, label=label)


def add_job(job: Job, graph: AGraph, props: Dict[str, Any], is_root: bool) -> None:
    """
    Adds a Job and its linked CallNode to the provided graph.

    Various combinations of properties will yield strictly CallNodes, strictly Jobs, or both.
    """
    call_node = job.call_node
    hash = f" {job.id[:8]}" if props["hash"] else ""
    opt_task = "" if props["detail"] else f"<BR />{job.call_node.task.fullname}"
    root = "ROOT<BR />" if is_root else ""

    if props["jobs"]:
        msg = f"{root}Job{hash}{opt_task}"
        if job.cached:
            msg = f'{msg}<BR /><FONT POINT-SIZE="10">(Cached)</FONT>'
        graph.add_node(job, fillcolor=JOB_COLOR, name=repr(job), label=f"<{msg}>")
    elif props["hash"]:
        props["job_str"] = f"Job{hash}\n"

    if props["detail"]:
        viz_call_node(call_node, graph, props)
        if props["jobs"]:
            graph.add_edge(job, call_node)


def process_routing_nodes(
    graph: AGraph,
    routing_arg: Any,
    routing_call_nodes: List[CallNode],
    props: Dict[str, Any],
) -> None:
    """
    Process a consecutive run of routing calls as a part of visualizing Values.

    After collecting the routing CallNodes that move around the same Value, we render them
    as subgraphs or nodes in the graph, depending on the properties.
    """
    add_value(routing_arg[0], routing_arg[1], graph, props)
    # Process consecutive runs of routing calls from outer to innermost
    routing_call_nodes.reverse()
    prev_graph = graph
    cluster_count = 0

    final_call_node = routing_call_nodes.pop(-1)
    if isinstance(final_call_node, CallNode):
        add_call_node(final_call_node, graph, props)
    graph.add_edge(routing_arg, final_call_node, color=DEFAULT_COLOR)

    for routing_call_node in routing_call_nodes:
        hash_str = f" {routing_call_node.call_hash[:8]}" if props["hash"] else ""
        label = f"CallNode{hash_str}\n{routing_call_node.task.fullname}"
        if props["wrap_calls"]:
            prev_graph.add_subgraph(
                [routing_arg],
                rank="same",
                name=f"cluster_{repr(routing_arg)}_{cluster_count}",
                margin=10.0,
                label=label,
                bgcolor=f"{CALL_NODE_COLOR}50",
            )
            prev_graph = graph.get_subgraph(f"cluster_{repr(routing_arg)}_{cluster_count}")
            cluster_count += 1
        else:
            add_call_node(routing_call_node, graph, props)
            graph.add_edge(routing_arg, routing_call_node, color=DEFAULT_COLOR)


def viz_value(scheduler: Scheduler, value: Value, graph: AGraph, props: Dict[str, Any]) -> None:
    """
    Visualizes a Value.

    Values are visualized by their upstream lineage.
    """
    routing_call_nodes = []
    routing_arg = None

    assert isinstance(scheduler.backend, RedunBackendDb)

    # Get upstream dataflow for value.
    dataflow_edges = walk_dataflow(scheduler.backend, value)

    for edge in dataflow_edges:
        src, dest = edge
        is_routing_edge, is_subvalue_edge = False, False
        if isinstance(src, CallNode):
            add_call_node(src, graph, props)
            complete_src: Any = src
        else:
            if isinstance(src, Value):
                # Then the upstream has to be a CallNode
                upstream: Any = [dest]
            elif isinstance(src, ArgumentValue):
                upstream = src.argument.upstream
                if not upstream:
                    upstream = src.argument.arg_hash
                src = src.value
                if not dest:
                    continue
                if isinstance(dest, ArgumentValue):
                    is_routing_edge = True
                if isinstance(dest, CallNodeValue):
                    is_subvalue_edge = True
                    # Handling special case of 1-to-1 mapping from CallNodeValue to ArgumentValue
                    # where there is no sub-value derivation involved
                    if dest.value == src:
                        continue
            elif isinstance(src, CallNodeValue):
                src, upstream = src.value, [src.call_node]
            else:
                raise ValueError(f"Unexpected dataflow node: {src}")

            complete_src = (src, upstream)
            is_derived_and_routed = (
                is_subvalue_edge and routing_arg and routing_arg[0] == complete_src[0]
            )
            if is_routing_edge or is_derived_and_routed:
                # We want to use the last ArgumentValue generated in a run of routing calls
                # since it includes the most complete upstream information, so we remove the node
                # if it was pre-emptively added, and store the outgoing CallNode to
                # reincorporate later
                if graph.has_node(complete_src):
                    out_edges = graph.out_edges([complete_src])
                    assert len(out_edges) == 1
                    outgoing_call_node = out_edges[0][1]
                    graph.remove_node(complete_src)
                    routing_call_nodes.append(outgoing_call_node)
                routing_arg = complete_src
            if not is_routing_edge:
                add_value(src, upstream, graph, props)

        if dest:
            if isinstance(dest, CallNode):
                add_call_node(dest, graph, props)
                complete_dest: Any = dest
            elif isinstance(dest, ArgumentValue):
                if is_routing_edge:
                    dest = dest.argument.call_node
                    routing_call_nodes.append(dest)
                else:
                    upstream = dest.argument.upstream
                    if not upstream:
                        upstream = dest.argument.arg_hash
                    dest = dest.value
                    add_value(dest, upstream, graph, props)
                    complete_dest = (dest, upstream)
            elif isinstance(dest, CallNodeValue):
                dest, upstream = dest.value, [dest.call_node]
                add_value(dest, upstream, graph, props)
                complete_dest = (dest, upstream)

            if not is_routing_edge:
                if routing_call_nodes:
                    process_routing_nodes(graph, routing_arg, routing_call_nodes, props)
                    routing_call_nodes = []
                color = SUBVALUE_EDGE_COLOR if is_subvalue_edge else DEFAULT_COLOR
                graph.add_edge(complete_dest, complete_src, color=color)

    if routing_call_nodes:
        process_routing_nodes(graph, routing_arg, routing_call_nodes, props)


def viz_call_node(call_node: CallNode, graph: AGraph, props: Dict[str, Any]) -> None:
    """
    Visualizes a CallNode.

    CallNodes are visualized from parents to children. CallNodes are the anchors
    by which argument and result values are also visualized.
    """
    add_call_node(call_node, graph, props)
    subgraph_nodes: List[Any] = [call_node]

    # Draw edges to parent CallNodes.
    if props.get("call_node_parents", True):
        for parent_call_node in call_node.parents:
            if parent_call_node.call_hash in props["valid_call_nodes"]:
                graph.add_edge(parent_call_node, call_node)

    # Draw argument nodes.
    for arg in call_node.arguments:
        arg_val = arg.value

        upstream = None
        if not props["deduplicate"]:
            upstream = arg.upstream
            # Guarantee node uniqueness for top level origin values with no upstream
            # by using their arg_hash instead
            if not upstream:
                upstream = arg.arg_hash

        values = add_value(arg_val, upstream, graph, props)
        for value in values:
            graph.add_edge(value, call_node)
            subgraph_nodes.append(value)

    # Draw the result node.
    result = call_node.value
    upstream = None if props["deduplicate"] else [call_node]
    values = add_value(result, upstream, graph, props)
    for value in values:
        graph.add_edge(call_node, value)
        subgraph_nodes.append(value)

    cluster_name = f"cluster-{call_node.call_hash[:8]}"
    graph.add_subgraph(subgraph_nodes, rank="same", name=cluster_name, style="invis")


def viz_job(job: Job, graph: AGraph, props: Dict[str, Any], is_root: bool) -> None:
    """
    Recursively visualizes a Job and all its children Jobs.
    """
    if is_root:
        add_job(job, graph, props, is_root)
        if props["jobs"]:
            graph.add_subgraph([job], rank="same", name="cluster-jobs", style="invis")
    for child_job in job.child_jobs:
        add_job(child_job, graph, props, False)
        if props["jobs"]:
            graph.add_edge(job, child_job)
            graph.get_subgraph("cluster-jobs").add_node(child_job)

        viz_job(child_job, graph, props, False)


def viz_execution(
    scheduler: Scheduler, execution: Execution, graph: AGraph, props: Dict[str, Any]
) -> None:
    """
    Visualizes an Execution.

    Executions are visualized by their root level job.
    """
    props["call_node_parents"] = True
    populate_valid_nodes(scheduler, execution.id, props)
    viz_job(execution.job, graph, props, True)


def viz_record(scheduler: Scheduler, record: Any, props: Dict[str, Any]) -> None:
    """
    Visualizes a record, generating resulting dot or png files as configured in the properties.

    Supports visualizing Executions, Jobs, CallNodes, and Values.
    Does NOT support task visualization.
    """
    # Override invalid flag combinations
    if not props["jobs"] and not props["detail"]:
        props["detail"] = True

    graph = init_graph(props["direction"])
    props["call_node_parents"] = False

    if isinstance(record, Execution):
        viz_execution(scheduler, record, graph, props)
        output = record.id
    elif isinstance(record, Job):
        add_job(record, graph, props, True)
        output = record.id
    elif isinstance(record, CallNode):
        viz_call_node(record, graph, props)
        output = record.call_hash
    elif isinstance(record, Value):
        viz_value(scheduler, record, graph, props)
        output = record.value_hash
    else:
        # We don't support Task visualization.
        raise ValueError(f"Visualization of this record type ({type(record)}) is not supported.")

    graph.layout(prog="dot")

    if props.get("output"):
        output = props["output"]

    format = props.get("format")
    if format == "png":
        graph.draw(f"{output}.png")
    elif format == "dot":
        graph.write(f"{output}.dot")
    elif format:
        raise ValueError("This visualization format is not supported.")
    else:
        graph.draw(f"{output}.png")
        graph.write(f"{output}.dot")



import argparse
import io
from argparse import Namespace
from typing import List, Optional
from unittest.mock import patch


def parse_args(parser: argparse.ArgumentParser, argv: List[str] = []) -> Namespace:
    """
    Parse an argv into an argparse.Namespace.
    """
    # argparse frequently tries to write to stderr, textual throws an exception
    # if anyone writes to stderr. We use a patch here to intercept writes to stderr.
    stderr = io.StringIO()
    with patch("sys.stderr", stderr):
        args, extra_args = parser.parse_known_args(argv)
    stderr.seek(0)
    return args


def get_parser_action(parser: argparse.ArgumentParser, key: str) -> Optional[argparse.Action]:
    """
    Returns the associated ArgumentParser Action for the given Namespace key.
    """
    for option in parser._option_string_actions.values():
        if option.dest == key:
            return option
    return None


def format_args(parser: argparse.ArgumentParser, args: Namespace) -> List[str]:
    """
    Format a Namespace back into an argv.
    """
    argv = []

    for key, value in args.__dict__.items():
        action = get_parser_action(parser, key)
        if isinstance(action, argparse._StoreAction) and value != action.default:
            argv.extend([action.option_strings[0], str(value)])

        elif isinstance(action, argparse._StoreTrueAction):
            if value:
                argv.append(action.option_strings[0])

        elif isinstance(action, argparse._AppendAction) and value != action.default:
            for val in value:
                argv.extend([action.option_strings[0], str(val)])

    return argv



import datetime
import pickle

import pytest
from dateutil.tz import tzoffset

from redun import File, Scheduler, task
from redun.tests.utils import use_tempdir
from redun.utils import pickle_dumps
from redun.value import (
    FileCache,
    Function,
    InvalidValueError,
    get_type_registry,
    is_unknown_function,
)


class Data:
    """
    Custom datatype.
    """

    def __init__(self, data):
        self.data = data


class DataType(FileCache):
    """
    Register Data to use file-based caching.
    """

    type = Data
    base_path = "tmp"

    def _serialize(self):
        # User defined serialization.
        return pickle_dumps(self.instance.data)

    @classmethod
    def _deserialize(cls, bytes):
        # User defined deserialization.
        return Data(pickle.loads(bytes))


@use_tempdir
def test_value_serialization() -> None:
    task_calls = []

    @task()
    def task1(data):
        task_calls.append("task1")
        return data

    @task()
    def main():
        data = Data("hello")
        return task1(data)

    scheduler = Scheduler()

    assert scheduler.run(main()).data == "hello"
    assert task_calls == ["task1"]

    # Value is cached to filesystem.
    assert File("tmp/773babcb7e0ba318c7981cb5595a5cbe640ab156").exists()

    assert scheduler.run(main()).data == "hello"
    assert task_calls == ["task1"]

    # Delete file.
    File("tmp/773babcb7e0ba318c7981cb5595a5cbe640ab156").remove()

    # Task will safely run again.
    assert scheduler.run(main()).data == "hello"
    assert task_calls == ["task1", "task1"]


def test_parse_datetime() -> None:
    """
    The TypeRegistry should be able to parse datetimes.
    """
    registry = get_type_registry()
    assert registry.parse_arg(datetime.datetime, "2022-07-01") == datetime.datetime(
        2022, 7, 1, 0, 0
    )
    assert registry.parse_arg(datetime.datetime, "2022-07-01 10:05:33") == datetime.datetime(
        2022, 7, 1, 10, 5, 33
    )
    assert registry.parse_arg(datetime.datetime, "2022-07-01 10:05:33 +0400") == datetime.datetime(
        2022, 7, 1, 10, 5, 33, tzinfo=tzoffset(None, 14400)
    )


def hello():
    # Example plain python function that we use below as an argument and return value in workflows.
    return "hello"


def test_function(scheduler: Scheduler) -> None:
    """
    Plain Python functions should be usable as arguments and return values.
    """

    @task
    def return_func():
        return hello

    @task
    def take_func(func):
        return func()

    @task
    def main():
        func = return_func()
        return take_func(func)

    assert scheduler.run(return_func()) == hello
    assert scheduler.run(take_func(hello)) == "hello"
    assert scheduler.run(main()) == "hello"

    registry = get_type_registry()
    assert registry.get_hash(hello) == "d57e3c02fb399ad72e41eba2d18054e7a5d6a196"


def test_function_parse_arg() -> None:
    """
    Functions should be parsable from the command line.
    """
    registry = get_type_registry()
    assert registry.parse_arg(type(hello), "redun.tests.test_value.hello") == hello

    with pytest.raises(ValueError, match="Unexpected format for function name: hello"):
        registry.parse_arg(type(hello), "hello")

    with pytest.raises(ValueError, match="Function not found: bad_module.hello"):
        registry.parse_arg(type(hello), "bad_module.hello")

    with pytest.raises(ValueError, match="Function not found: redun.tests.test_value.bad_func"):
        registry.parse_arg(type(hello), "redun.tests.test_value.bad_func")


def test_local_function(scheduler: Scheduler) -> None:
    """
    Local functions should be detected and rejected.
    """

    def hello2():
        return "hello"

    @task
    def return_func():
        return hello2

    @task
    def return_lambda():
        return lambda: 10

    with pytest.raises(InvalidValueError):
        scheduler.run(return_func())

    with pytest.raises(InvalidValueError):
        scheduler.run(return_lambda())


def test_deleted_function(scheduler: Scheduler) -> None:
    """
    Deleted functions should be safely detected.
    """

    value = Function(hello)
    data = value.serialize()

    # Alter the function name to something that does not exist.
    data2 = data.replace(b"hello", b"xxxxx")

    # Try to deserialize the function.
    func = Function.deserialize("builtin.function", data2)
    assert is_unknown_function(func)

    # Value should not be valid.
    assert not get_type_registry().is_valid(func)

    # Calling the function should trigger the stub function and fail.
    with pytest.raises(
        ValueError, match="Function 'redun.tests.test_value.xxxxx' cannot be found"
    ):
        func()



import copy
import os
import sys
from typing import Optional

import pytest

from redun import Scheduler, task
from redun.config import Config
from redun.federated_tasks import (
    federated_task,
    lambda_federated_task,
    launch_federated_task,
    rest_federated_task,
)
from redun.scheduler import SchedulerError
from redun.scheduler_config import postprocess_config
from redun.tests.utils import use_tempdir, wait_until


@task(namespace="redun_test")
def module_task(x, y=3, result_path: Optional[str] = None):
    result = x + y

    # Write to a path, so we can tell if a fire-and-forget run worked.
    if result_path is not None:
        with open(result_path, "w") as fp:
            fp.write(f"{result}")

    return result


@use_tempdir
def test_federated_task() -> None:
    # Use a process executor
    config_dict = {
        "federated_tasks.sample_task": {
            "executor": "process",
            "namespace": "redun_test",
            "task_name": "module_task",
            "load_module": "redun.tests.test_federated_tasks",
            "config_dir": os.path.join(
                os.path.dirname(__file__),
                "test_data",
                "federated_configs",
                "federated_task_config",
            ),
            "new_execution": "True",
        },
        "executors.process": {
            "type": "local",
            "mode": "process",
            "start_method": "spawn",
        },
    }

    config = Config(config_dict=config_dict)
    config = postprocess_config(config, config_dir=os.getcwd())

    scheduler = Scheduler(config)
    scheduler.load()

    federated = federated_task("sample_task", 3)
    assert 6 == scheduler.run(federated)
    federated = federated_task("sample_task", 3, y=8)
    assert 11 == scheduler.run(federated)
    federated = federated_task("sample_task", x=4, y=8)
    assert 12 == scheduler.run(federated)

    # Check error on incorrect task
    federated = federated_task("wrong_task", task_args=[3])
    with pytest.raises(AssertionError, match="Could not find the entrypoint `wrong_task`"):
        assert 6 == scheduler.run(federated)

    # Check behavior on missing executor
    config_dict_broken = copy.deepcopy(config_dict)
    config_dict_broken["federated_tasks.sample_task"]["executor"] = "bad_executor"

    config = Config(config_dict=config_dict_broken)
    config = postprocess_config(config, config_dir=os.getcwd())

    scheduler = Scheduler(config)
    scheduler.load()

    federated = federated_task("sample_task", task_args=[4])
    with pytest.raises(
        SchedulerError,
        match='Unknown executor "bad_executor"',
    ):
        scheduler.run(federated)

    # Check behavior on missing entrypoint configs
    config_dict_broken = copy.deepcopy(config_dict)
    del config_dict_broken["federated_tasks.sample_task"]["executor"]

    config = Config(config_dict=config_dict_broken)
    config = postprocess_config(config, config_dir=os.getcwd())

    scheduler = Scheduler(config)
    scheduler.load()

    federated = federated_task("sample_task", task_args=[4])
    with pytest.raises(
        AssertionError,
        match="Federated task entry `sample_task` does not have the "
        "required keys, missing `{'executor'}`",
    ):
        scheduler.run(federated)


@pytest.mark.skipif(
    (sys.version_info.major, sys.version_info.minor) == (3, 8),
    reason="Test fails on 3.8 due to 3.8-specific python bug.",
)
def test_lambda_federated_task(tmpdir) -> None:
    """Demonstrate that we can use the data from the lambda invocation to start a task."""
    config_dir = os.path.join(
        os.path.dirname(__file__),
        "test_data",
        "federated_configs",
        "federated_demo_config",
    )

    config = Config()
    config.read_path(os.path.join(config_dir, "redun.ini"))
    config = postprocess_config(config, config_dir=os.getcwd())

    scheduler = Scheduler(config)
    scheduler.load()

    result_path = os.path.join(tmpdir, "out.txt")

    # Run the invocation, which packages the inputs for us. Use the scratch redun defaults to.
    execution_id, invocation_data = scheduler.run(
        lambda_federated_task(
            config_name=config_dir,
            entrypoint="sample_task",
            lambda_function_name="fake-lambda-function",
            scratch_prefix="/tmp/redun",
            dryrun=True,
            x=8,
            y=9,
            result_path=result_path,
        )
    )

    assert invocation_data["config_name"].endswith(
        "redun/tests/test_data/federated_configs/federated_demo_config"
    )
    assert invocation_data["entrypoint"] == "sample_task"
    assert invocation_data["input_path"].endswith("input")
    assert invocation_data["execution_id"] == execution_id

    invocation_data["federated_config_path"] = invocation_data.pop("config_name")

    # Now launch the task. Local executors happen to work because the process pool is shut down
    # gently.
    launch_federated_task(**invocation_data)

    # The launched task is async to us, so let it finish.
    wait_until(lambda: os.path.isfile(result_path), timeout=5)

    with open(result_path, "r") as fp:
        result = fp.read()
    assert result == "17"


def test_rest_federated_task(tmpdir) -> None:
    """Demonstrate that we can use the data from the REST submission to start a task."""
    config_dir = os.path.join(
        os.path.dirname(__file__),
        "test_data",
        "federated_configs",
        "federated_demo_config",
    )

    config = Config()
    config.read_path(os.path.join(config_dir, "redun.ini"))
    config = postprocess_config(config, config_dir=os.getcwd())

    scheduler = Scheduler(config)
    scheduler.load()

    result_path = os.path.join(tmpdir, "out.txt")

    # Run the submission, which packages the inputs for us. Use the scratch redun defaults to.
    execution_id, rest_data = scheduler.run(
        rest_federated_task(
            config_name=config_dir,
            entrypoint="sample_task",
            url="fake",
            scratch_prefix="/tmp/redun",
            dryrun=True,
            x=2,
            y=9,
            result_path=result_path,
        )
    )

    assert rest_data["config_name"].endswith(
        "redun/tests/test_data/federated_configs/federated_demo_config"
    )
    assert rest_data["entrypoint"] == "sample_task"
    assert rest_data["input_path"].endswith("input")
    assert rest_data["execution_id"] == execution_id

    rest_data["federated_config_path"] = rest_data.pop("config_name")

    # Now launch the task. Local executors happen to work because the process pool is shut down
    # gently.
    launch_federated_task(**rest_data)

    # The launched task is async to us, so let it finish.
    wait_until(lambda: os.path.isfile(result_path), timeout=15)

    with open(result_path, "r") as fp:
        result = fp.read()
    assert result == "11"



from unittest.mock import Mock

import pytest

from redun import AliasExecutor, Scheduler
from redun.config import Config
from redun.executors.base import ExecutorError


def test_alias_executor(scheduler: Scheduler) -> None:
    config = Config({"batch": {"target": "default"}})
    executor = AliasExecutor("batch", scheduler, config["batch"])
    scheduler.executors["default"] = Mock()
    executor.submit(None)  # type: ignore # fake arguments for the mock

    # mypy doesn't understand the mock on the line below
    scheduler.executors["default"].submit.assert_called_with(None)  # type: ignore

    with pytest.raises(ExecutorError, match="Could not find executor `missing` from options"):
        config = Config({"batch": {"target": "missing"}})
        executor = AliasExecutor("batch", scheduler, config["batch"])
        executor.submit(None)  # type: ignore # fake arguments for the mock

    # Now try initializing with the
    scheduler.executors["default"].name = "default"
    executor = AliasExecutor("batch", scheduler, target=scheduler.executors["default"])

    scheduler.executors["default"].reset_mock()  # type: ignore
    executor.submit(None)  # type: ignore # fake arguments for the mock

    # mypy doesn't understand the mock on the line below
    scheduler.executors["default"].submit.assert_called_with(None)  # type: ignore

    with pytest.raises(
        AssertionError, match="Exactly one of `target` or `config` should be provided"
    ):
        AliasExecutor(
            "batch",
            scheduler,
            config=config["batch"],
            target=scheduler.executors["default"],
        )



import logging
import re
from itertools import chain
from typing import Any, Dict, List, Optional

from redun.backends.db import Argument, Execution, Job, Tag, Task, Value
from redun.tags import format_tag_value
from redun.utils import format_timestamp, trim_string

NULL = object()
logger = logging.getLogger("redun.console")
logger.setLevel(logging.INFO)
logger.addHandler(logging.FileHandler("/tmp/redun.log"))


def log_write(*args: Any) -> None:
    """
    Debugging function to use when developing with Textual.
    """
    logger.info(" ".join(map(str, args)))


def format_link(link_pattern: str, tags: Dict[str, Any]) -> Optional[str]:
    """
    Format a link pattern using a tag dictionary.
    """

    def replace(match: re.Match) -> str:
        # Trim {{ }} brackets.
        key_regex = match[0][2:-2]

        # Parse out key and optional regex.
        if ":" in key_regex:
            key, regex = key_regex.split(":", 1)
        else:
            key, regex = key_regex, ""

        # Fetch value
        value = tags.get(key, NULL)
        if value is NULL:
            raise KeyError(key)
        value = str(value)

        if regex:
            # If a regex is given, reformat the value according to the regex.
            match2 = re.match(regex, value)
            if match2:
                try:
                    # Try a named group first.
                    value = match2["val"]
                except IndexError:
                    value = match2[1]
            else:
                raise KeyError(key)
        return value

    try:
        # Replace every instance of '{{varible}}' and '{{variable:regex}}'.
        return re.sub(r"\{\{([^}]|\}[^}])+\}\}", replace, link_pattern)
    except KeyError:
        return None


def get_links(link_patterns: List[str], tags: List[Tag]) -> List[str]:
    """
    Get links from a list of link patterns and redun tags.
    """
    tags_dict: Dict[str, Any] = {tag.key: tag.value for tag in tags}
    return list(
        filter(
            None,
            [format_link(link_pattern, tags_dict) for link_pattern in link_patterns],
        )
    )


def format_tags(tags: List[Tag], max_length: int = 100, color="#9999cc") -> str:
    """
    Format a set of tags.
    """
    if not tags:
        return ""

    def format_tag_key_value(key: str, value: Any) -> str:
        key = trim_string(key, max_length=max_length)
        value_str = trim_string(format_tag_value(value), max_length=max_length)
        return f"[{color}]{key}={value_str}[/]"

    tags = sorted(tags, key=lambda tag: tag.key)

    return ", ".join(format_tag_key_value(tag.key, tag.value) for tag in tags)


def format_arguments(args: List[Argument]) -> str:
    """
    Display CallNode arguments.

    For example, if `args` has 2 positional and 1 keyword argument, we would
    display that as:

        'prog', 10, extra_file=File(path=prog.c, hash=763bc10f)
    """
    pos_args = sorted(
        [arg for arg in args if arg.arg_position is not None],
        key=lambda arg: arg.arg_position,
    )
    kw_args = sorted([arg for arg in args if arg.arg_key is not None], key=lambda arg: arg.arg_key)

    text = ", ".join(
        chain(
            (trim_string(repr(arg.value.preview)) for arg in pos_args),
            ("{}={}".format(arg.arg_key, trim_string(repr(arg.value.preview))) for arg in kw_args),
        )
    )
    return text


def format_job(job: Job) -> str:
    """
    Format a redun Job into a string representation.
    """
    args = format_arguments(job.call_node.arguments) if job.call_hash else ""
    return f"[bold]{job.task.fullname}[/][#999999]({args})[/]"


def format_traceback(job: Job) -> str:
    """
    Format the call stack from Execution down to the given Job.
    """

    # Determine job stack.
    job_stack = []
    current_job = job
    while current_job:
        job_stack.append(current_job)
        current_job = current_job.parent_job

    parts = ["[@click=screen.click_exec]Exec {exec}[/] > ".format(exec=job.execution.id[:8])]

    if len(job_stack) > 2:
        parts.append(
            "({num_jobs} {unit}) > ".format(
                num_jobs=len(job_stack) - 2,
                unit="Jobs" if len(job_stack) - 2 > 1 else "Job",
            )
        )
    if len(job_stack) > 1:
        parts.append(
            "[@click=screen.click_parent_job]Job {job_id} {task_name}[/] > ".format(
                job_id=job_stack[1].id[:8],
                task_name=job_stack[1].task.name,
            )
        )
    parts.append(
        "[bold]Job {job_id} {task_name}[/]".format(
            job_id=job_stack[0].id[:8],
            task_name=job_stack[0].task.name,
        )
    )
    if job.child_jobs:
        parts.append(f" > [@click=screen.children]{len(job.child_jobs)} child jobs[/]")

    return f"[bold]Traceback:[/b] {''.join(parts)}"


def style_status(status: str) -> str:
    """
    Returns styled text for a job/execution status.
    """
    status2style = {
        "DONE": "white on #55aa55",
        "FAILED": "white on #aa5555",
        "CACHED": "white on #5555aa",
        "RUN ": "black on #aaaa55",
    }
    if status in status2style:
        return f"[{status2style[status]}]{status.center(6)}[/]"
    else:
        return f"[{status.center(6)}]"


def format_record(record: Any) -> str:
    """
    Format a redun repo record (e.g. Execution, Job, etc) into a string.
    """
    if isinstance(record, Execution):
        return (
            f"Exec {record.id[:8]} {style_status(record.status)} "
            f"{format_timestamp(record.job.start_time)}"
            f"[[bold]{record.job.task.namespace or 'no namespace'}[/bold]] "
        )
    elif isinstance(record, Job):
        return (
            f"Job {record.id[:8]} {style_status(record.status)} "
            f"{format_timestamp(record.start_time)} "
        ) + format_job(record)
    elif isinstance(record, Task):
        return f"Task {record.hash[:8]} {record.fullname}"
    elif isinstance(record, Value):
        return f"Value {record.value_hash[:8]} {trim_string(repr(record.preview))}"
    else:
        return repr(record)
